'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var umi = require('@metaplex-foundation/umi');
var errors = require('./errors.cjs');
var helpers = require('./helpers.cjs');

/* eslint-disable no-bitwise */
const assertRange = (serializer, min, max, value) => {
  if (value < min || value > max) {
    throw new errors.NumberOutOfRangeError(serializer, min, max, value);
  }
};
const assertEnoughBytes = (serializer, bytes, expected) => {
  if (bytes.length === 0) {
    throw new errors.DeserializingEmptyBufferError(serializer);
  }
  if (bytes.length < expected) {
    throw new errors.NotEnoughBytesError(serializer, expected, bytes.length);
  }
};
const u8 = (options = {}) => ({
  description: options.description ?? 'u8',
  fixedSize: 1,
  maxSize: 1,
  serialize(value) {
    assertRange('u8', 0, Number('0xff'), value);
    const buffer = new ArrayBuffer(1);
    new DataView(buffer).setUint8(0, value);
    return new Uint8Array(buffer);
  },
  deserialize(bytes, offset = 0) {
    assertEnoughBytes('u8', bytes.slice(offset), 1);
    const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 1)));
    return [view.getUint8(0), offset + 1];
  }
});
const i8 = (options = {}) => ({
  description: options.description ?? 'i8',
  fixedSize: 1,
  maxSize: 1,
  serialize(value) {
    const half = Number('0x7f');
    assertRange('i8', -half - 1, half, value);
    const buffer = new ArrayBuffer(1);
    new DataView(buffer).setInt8(0, value);
    return new Uint8Array(buffer);
  },
  deserialize(bytes, offset = 0) {
    assertEnoughBytes('i8', bytes.slice(offset), 1);
    const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 1)));
    return [view.getInt8(0), offset + 1];
  }
});
const u16 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'u16(le)' : 'u16(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 2,
    maxSize: 2,
    serialize(value) {
      assertRange('u16', 0, Number('0xffff'), value);
      const buffer = new ArrayBuffer(2);
      new DataView(buffer).setUint16(0, value, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('u16', bytes.slice(offset), 2);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 2)));
      return [view.getUint16(0, littleEndian), offset + 2];
    }
  };
};
const i16 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'i16(le)' : 'i16(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 2,
    maxSize: 2,
    serialize(value) {
      const half = Number('0x7fff');
      assertRange('i16', -half - 1, half, value);
      const buffer = new ArrayBuffer(2);
      new DataView(buffer).setInt16(0, value, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('i16', bytes.slice(offset), 2);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 2)));
      return [view.getInt16(0, littleEndian), offset + 2];
    }
  };
};
const u32 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'u32(le)' : 'u32(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 4,
    maxSize: 4,
    serialize(value) {
      assertRange('u32', 0, Number('0xffffffff'), value);
      const buffer = new ArrayBuffer(4);
      new DataView(buffer).setUint32(0, value, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('u32', bytes.slice(offset), 4);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 4)));
      return [view.getUint32(0, littleEndian), offset + 4];
    }
  };
};
const i32 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'i32(le)' : 'i32(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 4,
    maxSize: 4,
    serialize(value) {
      const half = Number('0x7fffffff');
      assertRange('i32', -half - 1, half, value);
      const buffer = new ArrayBuffer(4);
      new DataView(buffer).setInt32(0, value, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('i32', bytes.slice(offset), 4);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 4)));
      return [view.getInt32(0, littleEndian), offset + 4];
    }
  };
};
const u64 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'u64(le)' : 'u64(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 8,
    maxSize: 8,
    serialize(value) {
      const valueBigInt = BigInt(value);
      assertRange('u64', 0, BigInt('0xffffffffffffffff'), valueBigInt);
      const buffer = new ArrayBuffer(8);
      new DataView(buffer).setBigUint64(0, valueBigInt, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('u64', bytes.slice(offset), 8);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 8)));
      return [view.getBigUint64(0, littleEndian), offset + 8];
    }
  };
};
const i64 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'i64(le)' : 'i64(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 8,
    maxSize: 8,
    serialize(value) {
      const valueBigInt = BigInt(value);
      const half = BigInt('0x7fffffffffffffff');
      assertRange('i64', -half - 1n, half, valueBigInt);
      const buffer = new ArrayBuffer(8);
      new DataView(buffer).setBigInt64(0, valueBigInt, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('i64', bytes.slice(offset), 8);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 8)));
      return [view.getBigInt64(0, littleEndian), offset + 8];
    }
  };
};
const u128 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'u128(le)' : 'u128(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 16,
    maxSize: 16,
    serialize(value) {
      const valueBigInt = BigInt(value);
      const max = BigInt('0xffffffffffffffffffffffffffffffff');
      assertRange('u128', 0, max, valueBigInt);
      const buffer = new ArrayBuffer(16);
      const view = new DataView(buffer);
      const leftOffset = littleEndian ? 8 : 0;
      const rightOffset = littleEndian ? 0 : 8;
      const rightMask = 0xffffffffffffffffn;
      view.setBigUint64(leftOffset, valueBigInt >> 64n, littleEndian);
      view.setBigUint64(rightOffset, valueBigInt & rightMask, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('u128', bytes.slice(offset), 16);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 16)));
      const leftOffset = littleEndian ? 8 : 0;
      const rightOffset = littleEndian ? 0 : 8;
      const left = view.getBigUint64(leftOffset, littleEndian);
      const right = view.getBigUint64(rightOffset, littleEndian);
      return [(left << 64n) + right, offset + 16];
    }
  };
};
const i128 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'i128(le)' : 'i128(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 16,
    maxSize: 16,
    serialize(value) {
      const valueBigInt = BigInt(value);
      const half = BigInt('0x7fffffffffffffffffffffffffffffff');
      assertRange('i128', -half - 1n, half, valueBigInt);
      const buffer = new ArrayBuffer(16);
      const view = new DataView(buffer);
      const leftOffset = littleEndian ? 8 : 0;
      const rightOffset = littleEndian ? 0 : 8;
      const rightMask = 0xffffffffffffffffn;
      view.setBigInt64(leftOffset, valueBigInt >> 64n, littleEndian);
      view.setBigUint64(rightOffset, valueBigInt & rightMask, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('i128', bytes.slice(offset), 16);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 16)));
      const leftOffset = littleEndian ? 8 : 0;
      const rightOffset = littleEndian ? 0 : 8;
      const left = view.getBigInt64(leftOffset, littleEndian);
      const right = view.getBigUint64(rightOffset, littleEndian);
      return [(left << 64n) + right, offset + 16];
    }
  };
};
const f32 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'f32(le)' : 'f32(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 4,
    maxSize: 4,
    serialize(value) {
      const buffer = new ArrayBuffer(4);
      new DataView(buffer).setFloat32(0, value, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('f32', bytes.slice(offset), 4);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 4)));
      return [view.getFloat32(0, littleEndian), offset + 4];
    }
  };
};
const f64 = (options = {}) => {
  const littleEndian = (options.endian ?? umi.Endian.Little) === umi.Endian.Little;
  const defaultDescription = littleEndian ? 'f64(le)' : 'f64(be)';
  return {
    description: options.description ?? defaultDescription,
    fixedSize: 8,
    maxSize: 8,
    serialize(value) {
      const buffer = new ArrayBuffer(8);
      new DataView(buffer).setFloat64(0, value, littleEndian);
      return new Uint8Array(buffer);
    },
    deserialize(bytes, offset = 0) {
      assertEnoughBytes('f64', bytes.slice(offset), 8);
      const view = new DataView(helpers.UInt8ArrayToBuffer(bytes.slice(offset, offset + 8)));
      return [view.getFloat64(0, littleEndian), offset + 8];
    }
  };
};

exports.f32 = f32;
exports.f64 = f64;
exports.i128 = i128;
exports.i16 = i16;
exports.i32 = i32;
exports.i64 = i64;
exports.i8 = i8;
exports.u128 = u128;
exports.u16 = u16;
exports.u32 = u32;
exports.u64 = u64;
exports.u8 = u8;
//# sourceMappingURL=numbers.cjs.map
